# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for ctrl_bps_panda utilities.
"""
import unittest

from lsst.ctrl.bps import GenericWorkflowExec, GenericWorkflowJob
from lsst.ctrl.bps.panda.constants import PANDA_MAX_LEN_INPUT_FILE
from lsst.ctrl.bps.panda.utils import _make_pseudo_filename


class TestPandaUtils(unittest.TestCase):
    """Simple test of utilities."""

    def testTooLongPseudoFilename(self):
        # define enough of a job for this test
        myexec = GenericWorkflowExec("test_exec")
        myexec.src_uri = "/dummy/path/test_exec"
        gwjob = GenericWorkflowJob("j" * PANDA_MAX_LEN_INPUT_FILE)
        gwjob.executable = myexec
        gwjob.arguments = ""
        with self.assertRaises(RuntimeError):
            _ = _make_pseudo_filename({}, gwjob)

    def testOKPseudoFilename(self):
        # define enough of a job for this test
        myexec = GenericWorkflowExec("test_exec")
        myexec.src_uri = "/dummy/path/test_exec"
        gwjob = GenericWorkflowJob("j" * 15)
        gwjob.executable = myexec
        gwjob.arguments = ""
        name = _make_pseudo_filename({}, gwjob)
        self.assertIn("j" * 15, name)


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import re

from lsst.ctrl.bps.panda.constants import PANDA_MAX_LEN_INPUT_FILE

_LOG = logging.getLogger(__name__)


class CommandLineEmbedder:
    """Class embeds static (constant across a task) values
    into the pipeline execution command line
    and resolves submission side environment variables

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration that includes the list of dynamic
        (uniques per job) and submission side resolved variables
    """

    def __init__(self, config):
        self.leave_placeholder_params = config.get("placeholderParams", ["qgraphNodeId", "qgraphId"])
        self.submit_side_resolved = config.get("submitSideResolvedParams", ["USER"])

    def replace_static_parameters(self, cmd_line, lazy_vars):
        """Substitutes the lazy parameters in the command line which
        are static, the same for every job in the workflow and could be
        defined once.

        This function offloads the edge node processing
        and number of parameters transferred together with job

        Parameters
        ----------
        cmd_line:  `str`
            Command line to be processed.
        lazy_vars : `dict`
            Lazy variables and its values.

        Returns
        -------
        cmd : `str`
            Processed command line.
        """
        for param_name, param_val in lazy_vars.items():
            if param_name not in self.leave_placeholder_params:
                cmd_line = cmd_line.replace("{" + param_name + "}", param_val)
        return cmd_line

    def replace_static_files(self, cmd_line, files):
        """Substitute the FILE keys with values in the command line
        which are static, the same for every job in the workflow and
        could be defined once.

        Parameters
        ----------
        cmd_line: `str`
            command line to be processed
        files: `list` [`lsst.ctrl.bps.GenericWorkflowFile`]
            input and output files for the job.

        Returns
        -------
        cmd : `str`
            Processed command line.
        """
        # make copy of given command line for error message.
        orig_cmd_line = cmd_line

        # make gwfile lookup by name
        files_by_name = {}
        for gwfile in files:
            files_by_name[gwfile.name] = gwfile

        for file_key in re.findall(r"<FILE:([^>]+)>", cmd_line):
            try:
                gwfile = files_by_name[file_key]
            except KeyError as e:
                raise RuntimeError(
                    "%s in command line, but corresponding file not given to function (%s)",
                    file_key,
                    orig_cmd_line,
                ) from e

            if not gwfile.wms_transfer and gwfile.job_access_remote:
                cmd_line = cmd_line.replace(f"<FILE:{gwfile.name}>", gwfile.src_uri)
        return cmd_line

    def resolve_submission_side_env_vars(self, cmd_line):
        """Substitute the lazy parameters in the command line
        which are defined and resolved on the submission side.

        Parameters
        ----------
        cmd_line : `str`
            Command line to be processed.

        Returns
        -------
        cmd : `str`
            Processed command line.
        """
        for param in self.submit_side_resolved:
            if os.getenv(param):
                cmd_line = cmd_line.replace("<ENV:" + param + ">", os.getenv(param))
            else:
                _LOG.info("Expected parameter %s is not found in the environment variables", param)
        return cmd_line

    def attach_pseudo_file_params(self, lazy_vars):
        """Add the parameters needed to finalize creation of a pseudo file.

        Parameters
        ----------
        lazy_vars : `dict`
            Values to be substituted.

        Returns
        -------
        suffix : `str`
            Pseudo input file name suffix.
        """
        file_suffix = ""
        for item in self.leave_placeholder_params:
            file_suffix += "+" + item + ":" + lazy_vars.get(item, "")
        return file_suffix

    def substitute_command_line(self, cmd_line, lazy_vars, job_name, gwfiles):
        """Preprocess the command line leaving for the edge node evaluation
        only parameters which are job / environment dependent

        Parameters
        ----------
        cmd_line: `str`
            Command line containing all lazy placeholders.
        lazy_vars: `dict` [ `str`, `str` ]
            Lazy parameter name/values.
        job_name: `str`
            Job name proposed by BPS.
        gwfiles: `list` [`lsst.ctrl.bps.GenericWorkflowFile`]
            Job files.

        Returns
        -------
        cmd_line: `str`
            processed command line
        file_name: `str`
            job pseudo input file name

        Raises
        ------
        RuntimeError
            Raised if pseudo input filename is too long.
        """
        cmd_vals = {m.group(1) for m in re.finditer(r"[^$]{([^}]+)}", cmd_line)}
        actual_lazy_vars = {}
        for key in cmd_vals:
            actual_lazy_vars[key] = lazy_vars[key]

        cmd_line = self.replace_static_parameters(cmd_line, actual_lazy_vars)
        cmd_line = self.resolve_submission_side_env_vars(cmd_line)
        if gwfiles:
            cmd_line = self.replace_static_files(cmd_line, gwfiles)
        file_name = job_name + self.attach_pseudo_file_params(actual_lazy_vars)

        if len(file_name) > PANDA_MAX_LEN_INPUT_FILE:
            _LOG.error(f"Too long pseudo input filename: {file_name}")
            raise RuntimeError(
                f"job pseudo input file name contains more than {PANDA_MAX_LEN_INPUT_FILE} symbols. Aborting."
            )

        return cmd_line, file_name



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for PanDA authentication utilities.
"""
import os
import unittest
from unittest import mock

from lsst.ctrl.bps.panda import __version__ as version
from lsst.ctrl.bps.panda.panda_auth_utils import panda_auth_status


class VersionTestCase(unittest.TestCase):
    """Test versioning."""

    def test_version(self):
        # Check that version is defined.
        self.assertIsNotNone(version)


class TestPandaAuthUtils(unittest.TestCase):
    """Simple test of auth utilities."""

    def testPandaAuthStatusWrongEnviron(self):
        unwanted = {
            "PANDA_AUTH",
            "PANDA_VERIFY_HOST",
            "PANDA_AUTH_VO",
            "PANDA_URL_SSL",
            "PANDA_URL",
        }
        test_environ = {key: val for key, val in os.environ.items() if key not in unwanted}
        with mock.patch.dict(os.environ, test_environ, clear=True):
            with self.assertRaises(OSError):
                panda_auth_status()


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for cmd_line_embedder.py
"""
import unittest

from lsst.ctrl.bps import GenericWorkflowFile
from lsst.ctrl.bps.panda.cmd_line_embedder import CommandLineEmbedder
from lsst.ctrl.bps.panda.constants import PANDA_MAX_LEN_INPUT_FILE


class TestCmdLineEmbedder(unittest.TestCase):
    """Test CmdLineEmbedder class"""

    def setUp(self):
        self.cmd_line_0 = "<ENV:CTRL_MPEXEC_DIR>/mycmd -b {key1} -c {key2} -d {key3} {key4}"
        self.cmd_line_1 = "<ENV:CTRL_MPEXEC_DIR>/mycmd -b <FILE:file3> -c <FILE:file4> -d {key3} {key4}"
        self.cmd_line_2 = (
            "<ENV:CTRL_MPEXEC_DIR>/mycmd -b <FILE:sfile1> -c <FILE:file3> -d <FILE:sfile2> <FILE:file4>"
        )
        self.sfile1 = GenericWorkflowFile("sfile1", "/path/to/sfile1.yaml", False, True, True)
        self.sfile2 = GenericWorkflowFile("sfile2", "/path/to/sfile2.yaml", False, True, True)
        self.file3 = GenericWorkflowFile("file3", "/path/to/file3.yaml", True, False, False)
        self.file4 = GenericWorkflowFile("file4", "/path/to/file4.yaml", True, False, False)

        self.ans_cmd_line_0 = "<ENV:CTRL_MPEXEC_DIR>/mycmd -b {key1} -c {key2} -d {key3} {key4}"
        self.ans_cmd_line_1 = "<ENV:CTRL_MPEXEC_DIR>/mycmd -b <FILE:file3> -c <FILE:file4> -d {key3} {key4}"
        self.ans_cmd_line_2 = (
            "<ENV:CTRL_MPEXEC_DIR>/mycmd -b /path/to/sfile1.yaml -c <FILE:file3> "
            "-d /path/to/sfile2.yaml <FILE:file4>"
        )

    def testReplaceStaticFilesNoFiles(self):
        orig_cmd_line = self.cmd_line_0
        orig_cmd_line_copy = orig_cmd_line

        cmd_line_embedder = CommandLineEmbedder({})
        new_cmd_line = cmd_line_embedder.replace_static_files(orig_cmd_line, [])

        # Ensure no side effect
        self.assertEqual(orig_cmd_line, orig_cmd_line_copy)
        self.assertEqual(new_cmd_line, orig_cmd_line_copy)

    def testReplaceStaticFilesMissingFiles(self):
        orig_cmd_line = self.cmd_line_1
        orig_cmd_line_copy = orig_cmd_line

        cmd_line_embedder = CommandLineEmbedder({})
        with self.assertRaises(RuntimeError):
            _ = cmd_line_embedder.replace_static_files(orig_cmd_line, [self.sfile1, self.file3])

        self.assertEqual(orig_cmd_line, orig_cmd_line_copy)

    def testReplaceStaticFilesNone(self):
        orig_cmd_line = self.cmd_line_1
        orig_cmd_line_copy = orig_cmd_line

        cmd_line_embedder = CommandLineEmbedder({})
        new_cmd_line = cmd_line_embedder.replace_static_files(
            orig_cmd_line, [self.sfile1, self.sfile2, self.file3, self.file4]
        )

        self.assertEqual(orig_cmd_line, orig_cmd_line_copy)
        self.assertEqual(new_cmd_line, self.ans_cmd_line_1)

    def testReplaceStaticFilesSome(self):
        orig_cmd_line = self.cmd_line_2
        orig_cmd_line_copy = orig_cmd_line

        cmd_line_embedder = CommandLineEmbedder({})
        new_cmd_line = cmd_line_embedder.replace_static_files(
            orig_cmd_line, [self.sfile1, self.sfile2, self.file3, self.file4]
        )

        self.assertEqual(orig_cmd_line, orig_cmd_line_copy)
        self.assertEqual(new_cmd_line, self.ans_cmd_line_2)

    def testTooLongPseudoFilename(self):
        cmd_line_embedder = CommandLineEmbedder({})
        with self.assertRaises(RuntimeError):
            _, _ = cmd_line_embedder.substitute_command_line("", {}, "j" * PANDA_MAX_LEN_INPUT_FILE, [])

    def testOKPseudoFilename(self):
        cmd_line_embedder = CommandLineEmbedder({})
        _, name = cmd_line_embedder.substitute_command_line("", {}, "j" * 15, [])
        self.assertIn("j" * 15, name)


if __name__ == "__main__":
    unittest.main()
